@@ -1,16 +1,19 @@
 
+import re
 import os
 import sys
 import json
 import pickle
+import humanize
 import datetime
 import inspect
 import logging
 from time import time
 from functools import wraps
+from datetime import datetime
 from typing import Tuple, Callable, Union
 from .utils import get_params, parse_time, random_string
-from .backends import get_load_dump_from_path
+from .backends import Backend
 
 # Dictionary are not hashable and the python hash is not consistent
 # between runs so we have to use an external dictionary hashing package
@@ -43,6 +46,9 @@ class Cache:
         log_level: str = "critical",
         log_format: str = '%(asctime)-15s [%(levelname)s]: %(message)s',
         backup_path: str = None,
+        backup: bool = True,
+        dump_kwargs:dict = {},
+        load_kwargs:dict = {},
     ):
         """
         Cache the results of a function (or method).
@@ -137,6 +143,9 @@ class Cache:
             For this reason in the formatter you can use any variable such as {cache_dir}, {cache_path}, or the arguments
             of the function. Moreover, there is also another two additional parameters, {_date} which is the date of the backup, and {_rnd} which is a random string that will 
             guarantee that no file has the same name.  
+        backup: bool = True,
+            If the cache should backup the result to a .pkl in case of exception during the serializzation.
+            This flag is mainly for debug pourpouses.
             ```
         """
         self.log_level = log_level
@@ -145,12 +154,13 @@ class Cache:
         self.use_source_code = use_source_code
         self.validity_duration = parse_time(validity_duration)
 
-        self._load, self._dump = get_load_dump_from_path(cache_path)
-
         self.cache_path = cache_path
-        self.backup_path = backup_path or (cache_path + "_backup.pkl")
+        self.is_backup_enabled = backup
+        self.backup_path = backup_path
         self.cache_dir = cache_dir or os.environ.get("CACHE_DIR", "./cache")
 
+        self.load_kwargs, self.dump_kwargs = load_kwargs, dump_kwargs
+
     @staticmethod
     def store(obj, path: str) -> None:
         """Store an object at a path, this automatically choose the correct backend.
@@ -162,9 +172,8 @@ class Cache:
             path: str,
                 Where to store the file, based on its extension it will choose the correct backend.
         """
-        load, dump = get_load_dump_from_path(path)
         os.makedirs(os.path.dirname(os.path.abspath(path)), exist_ok=True)
-        dump(obj, path)
+        Backend({}, {}).dump(obj, path)
 
     @staticmethod
     def load(path: str):
@@ -180,8 +189,7 @@ class Cache:
         -------
         The loaded object.
         """
-        load, dump = get_load_dump_from_path(path)
-        return load(path)
+        return Backend({}, {}).load({}, path)
 
     @staticmethod
     def compute_path(function: Callable, *args, **kwargs) -> str:
@@ -196,7 +204,7 @@ class Cache:
         instance = getattr(function, "__cacher_instance")
         function = getattr(function, "__cached_function")
 
-        return instance._get_formatted_path(args, kwargs, instance._compute_function_info(function))
+        return instance._get_formatted_path(args, kwargs, function_info=instance._compute_function_info(function))
 
 
     def _compute_function_info(self, function: Callable):
@@ -224,14 +232,27 @@ class Cache:
 
     def _backup(self, result, path, exception, args, kwargs):
         """This function handle the backupping of the data when an the serialization fails."""
-        date = datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
+
+        # Check if it's a structured path
+        if isinstance(path, list) or isinstance(path, tuple):
+            return self._backup(result, path[0], exception, args, kwargs)
+                
+        elif isinstance(path, dict):
+            return self._backup(result, next(path.values()), exception, args, kwargs)
+
+        date = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
         # Ensure that the we won't overwrite anything
         while True:
             # Generate a new random value
             rnd = random_string(40) # hardcoded length but if they are enough for git it's enough for us.
             backup_path = self._get_formatted_path(
                 args, kwargs, 
-                backup=True, extra_kwargs={"_date":date, "_rnd":rnd}
+                formatter=self.backup_path,
+                extra_kwargs={
+                    "_date":date, 
+                    "_rnd":rnd,
+                    "cache_path":self._get_formatted_path(args, kwargs),
+                }
             )
             # Check for the existance
             if os.path.exists(backup_path):
@@ -255,7 +276,143 @@ class Cache:
             pickle.dump(result, f)
         # Re-raise the exception
         exception.backup_path = backup_path
-        raise exception
+        exception.path = path
+        exception.result = result
+        return exception
+
+    def _get_metadata_path(self, path):
+        return path + ".metadata"
+
+    def _load(self, path):
+
+        # Check if it's a structured path
+        if isinstance(path, list) or isinstance(path, tuple):
+            result = []
+            for p in path:
+                cache = self._load(p)
+
+                if cache is None:
+                    return None
+
+                result.append(cache)
+
+            if isinstance(path, tuple):
+                result = tuple(result)
+            return result
+
+        elif isinstance(path, dict):
+            result = {}
+            for key, p in path.items():
+                cache = self._load(p)
+
+                if cache is None:
+                    return None
+
+                result[key] = cache
+            return result
+
+        # Check if the cache exists and is readable
+        if not os.path.isfile(path):
+            self.logger.info("The cache at path '%s' does not exists.", path)
+            return None
+
+        self.logger.info("Loading cache from %s", path)
+
+        metadata_path = self._get_metadata_path(path)
+
+        # Load the metadata if present
+        if os.path.isfile(metadata_path):
+            self.logger.info("Loading the metadata file at '%s'", metadata_path)
+            with open(metadata_path, "r") as f:
+                metadata = json.load(f)
+        else:
+            self.logger.info("The metadata file at '%s' do not exists.", metadata_path)
+            # TODO: do we need to to more stuff?
+            metadata = {}
+
+        # Check if the cache is still valid
+        if self.validity_duration is not None:
+            enlapsed_time = time() - metadata.get("creation_time", float("-inf"))
+            if  enlapsed_time > self.validity_duration:
+                os.remove(path)
+                return None 
+
+        # actually load the values
+        return Backend(self.load_kwargs, self.dump_kwargs).load(metadata.get("backend_metadata", {}), path)
+
+    def _dump(self, args, kwargs, result, path, start_time, end_time):
+
+        # Check if it's a structured path
+        if isinstance(path, list) or isinstance(path, tuple):
+            assert isinstance(result,list) or isinstance(result,tuple)
+            assert len(result) == len(path)
+            for r, p in zip(result, path):
+                self._dump(args, kwargs, r, p, start_time, end_time)
+            return 
+                
+        elif isinstance(path, dict):
+            assert isinstance(result, dict)
+            assert set(result.keys()) == set(path.keys())
+
+            for key in result.keys():
+                self._dump(args, kwargs, result[key], path[key], start_time, end_time)
+            return 
+        
+
+        # Dump the file
+        self.logger.info("Saving the cache at %s", path)
+        os.makedirs(os.path.dirname(path), exist_ok=True)
+        dump_start_time = time()
+        backend_metadata = Backend(self.load_kwargs, self.dump_kwargs).dump(result, path) or {}
+        dump_end_time = time()
+
+        # Compute the metadata
+        metadata = {
+            # When the cache was created
+            "creation_time": start_time,
+            "creation_time_human": datetime.fromtimestamp(
+                start_time
+            ).strftime("%Y-%m-%d %H:%M:%S"),
+
+            # How much the function took to compute the result
+            "time_delta":end_time - start_time,
+            "time_delta_human":humanize.precisedelta(end_time - start_time),
+
+            # How much time it took to serialize the result and save it to a file
+            "file_dump_time":dump_end_time - dump_start_time,
+            "file_dump_time_human":humanize.precisedelta(
+                dump_end_time - dump_start_time
+            ),
+
+            # How big is the serialized result
+            "file_dump_size":os.path.getsize(path),
+            "file_dump_size_human":humanize.naturalsize(os.path.getsize(path)),
+
+            # The arguments used to load and dump the file
+            "load_kwargs":self.load_kwargs,
+            "dump_kwargs":self.dump_kwargs,
+
+            # Informations about the function
+            "function_name":self.function_info["function_name"],
+            "function_file":"%s:%s"%(
+                self.decorated_function.__code__.co_filename,
+                self.decorated_function.__code__.co_firstlineno
+            ),
+            # TODO re-insert
+            #"parameter":get_params(self.function_info, args, kwargs),
+            "args_to_ignore":self.function_info["args_to_ignore"],
+            "source":self.function_info.get("source", None),
+
+            # The data reserved for the backend to corretly serialize and 
+            # de-serialize the values
+            "backend_metadata":backend_metadata,
+        }
+
+        metadata_path = self._get_metadata_path(path)
+        self.logger.info("Saving the cache meta-data at %s", metadata_path)
+        with open(metadata_path, "w") as f:
+            json.dump(metadata, f, indent=4)
+
 
     def _decorate_callable(self, function: Callable) -> Callable:
         # wraps to support pickling
@@ -263,25 +420,26 @@ class Cache:
         def wrapped(*args, **kwargs):
             # Get the path
             path = self._get_formatted_path(args, kwargs)
-            # ensure that the cache folder exist
-            os.makedirs(os.path.dirname(path), exist_ok=True)
-            # If the file exist, load it
-            if os.path.exists(path) and self._is_valid(path):
-                self.logger.info("Loading cache from {}".format(path))
-                return self._load(path)
-            # else call the function
+            # Try to load the cache
+            result = self._load(path)
+            # if we got a result, reutrn it
+            if result is not None:
+                return result
+            
+            self.logger.info("Computing the result for %s %s", args, kwargs)
+            # otherwise compute the result
+            start_time = time()
             result = function(*args, **kwargs)
-            # and save the result
-            self.logger.info("Saving the computed result at %s", path)
+            end_time = time()
+
+            # Save the result
             try:
-                self._dump(result, path)
+                self._dump(args, kwargs, result, path, start_time, end_time)
             except Exception as e:
-                self._backup(result, path, e, args, kwargs)
+                if self.is_backup_enabled:
+                    raise self._backup(result, path, e, args, kwargs)
+                raise e
 
-            # If the cache is supposed to have a
-            # validity duration then save the creation timestamp
-            if self.validity_duration:
-                self._save_creation_time(path)
             return result
 
         # add a reference to the cached function so we can unpack
@@ -290,43 +448,30 @@ class Cache:
         setattr(wrapped, "__cacher_instance", self)
         return wrapped
 
-    def _save_creation_time(self, path):
-        cache_date = path + "_time.json"
-        self.logger.info("Saving the cache time meta-data at %s", cache_date)
-        with open(cache_date, "w") as f:
-            json.dump({"creation_time": time()}, f)
+    def _get_formatted_path(self, args, kwargs, formatter=None, function_info=None, extra_kwargs=None) -> str:
+        """Compute the path adding and computing the needed arguments."""        
+        formatter = formatter or self.cache_path
 
-    def _is_valid(self, path):
-        # If validation to "" or 0
-        # then it's disabled and the cache is always valid
-        if not self.validity_duration:
-            return True
-        # path of the saved creation_time
-        date_path = path + "_time.json"
-        # Check if there is the creation_time
-        if not os.path.exists(date_path):
-            # in this case the cache file exists
-            # but not the creation time
-            # this might means that the file was deleted
-            # or the cache was previously used without
-            # validity time
-            self.logger.warning(
-                "Warning no creation time at %s. Therefore the cache will be considered not valid", date_path)
-            return False
-        # Open the file e confront the time
-        with open(date_path, "r") as f:
-            cache_time = json.load(f)["creation_time"]
-        return time() - cache_time < self.validity_duration
+        if isinstance(formatter, list):
+            return [
+                self._get_formatted_path(args, kwargs, f)
+                for f in formatter
+            ]
+
+        if isinstance(formatter, tuple):
+            return tuple([
+                self._get_formatted_path(args, kwargs, f)
+                for f in formatter
+            ])
+
+        elif isinstance(formatter, dict):
+            return {
+                key:self._get_formatted_path(args, kwargs, v)
+                for key, v in formatter.items()
+            }
 
-    def _get_formatted_path(self, args, kwargs, function_info=None, backup=False, extra_kwargs=None) -> str:
-        """Compute the path adding and computing the needed arguments."""
         extra_kwargs = extra_kwargs or {}
-        if backup:
-            formatter = self.backup_path
-            extra_kwargs["cache_path"] = self._get_formatted_path(args, kwargs)
-        else:
-            formatter = self.cache_path
-        
+
         function_info = function_info or self.function_info
         params = get_params(function_info, args, kwargs)
         
@@ -335,6 +480,20 @@ class Cache:
 
         self.logger.debug("Got parameters %s", params)
 
+        # Handle the composite paths
+        for match in re.finditer(r"\{([^\{]+)(:?\..+?)+\}", formatter):
+            # Extract the matching string
+            match = match.group(0)
+            # Get the name of the base element and the attributes chain
+            root, *attrs = match[1:-1].split(".")
+            # Get the params to use for the attributes chain
+            root = params[root]
+            # Follow the attributes chain
+            for attr in attrs:
+                root = getattr(root, attr)
+            # Replace the result in the formatter
+            formatter = formatter.replace(match, str(root))
+
         # Compute the path of the cache for these parameters
         path = formatter.format(
             cache_dir=self.cache_dir,
@@ -343,6 +502,7 @@ class Cache:
             **extra_kwargs,
         )
         self.logger.debug("Calculated path %s", path)
+
         return path
 
     def _fix_docs(self, function: Callable, wrapped: Callable) -> Callable:
@@ -354,15 +514,19 @@ class Cache:
 
     def decorate(self, function: Callable) -> Callable:
         self.function_info = self._compute_function_info(function)
+        self.decorated_function = function
         wrapped = self._decorate_callable(function)
         wrapped = self._fix_docs(function, wrapped)
         return wrapped
 
     def __call__(self, function):
         self.logger = logging.getLogger(__name__ + "." + function.__name__)
-        handler = logging.StreamHandler(sys.stderr)
-        handler.setFormatter(logging.Formatter(self.log_format))
-        self.logger.addHandler(handler)
+        # Do not re-initialize loggers if we have to cache multiple functions
+        # with the same name
+        if not self.logger.hasHandlers():
+            handler = logging.StreamHandler(sys.stderr)
+            handler.setFormatter(logging.Formatter(self.log_format))
+            self.logger.addHandler(handler)
 
         if self.log_level.lower() not in log_levels:
             raise ValueError("The logger level {} is not a supported one. The available ones are {}".format(
